#===============================================================================
# Constants (DO NOT EDIT THESE)
#===============================================================================

AVAILABLE_PROTOCOLS = (1,)

#===============================================================================
# Public API
#===============================================================================
class InvalidProtocol(Exception): pass


class Application(object):
    def __init__(self, api_key, server_key, app_key, protocol=1):
        """
        Set up a new application with a api_key and server_key to get information
        from the web and an app_name to identify it.
        """
        self.api_key = api_key
        self.server_key = server_key
        self.app_key = app_key
        protocol = int(protocol)
        if protocol not in AVAILABLE_PROTOCOLS:
            raise InvalidProcotol, "The protocol version '%s' is not available" % protocol
        self.protocol = protocol
        
    @property
    def _url(self):
        return 'http://lotteries.ojii.ch/api/%s/get_queue/%s/%s/%s/' % (self.protocol, self.api_key, self.server_key, self.app_key)
    
    def get_winners(self, callback=None, async=True):
        """
        Calls 'callback' with a list of steamid's that won since the last check
        """
        downloader = get_downloader()
        if async:
            if not callable(callback):
                raise TypeError, "'callback' argument in Application.get_winners must be callable if asynchronous flag is True"
            dl = downloader(self._url, callback, self.protocol)
        else:
            dl = downloader(self._url, callback, self.protocol)
        return dl.run(async)
    
    
def set_downloader(downloader):
    """
    Set a custom downloader. The downloader should be a subclass of
    BaseDownloader.
    """
    manager.set_downloader(downloader)
    

#===============================================================================
# Downloaders
#===============================================================================

class BaseDownloader(object):
    def __init__(self, url, callback, protocol_version):
        self.url = url
        self.callback = callback
        self.response_builder = get_response_builder(protocol_version)
        
    def run(self, async):
        if async:
            return self.handle_async()
        else:
            return self.handle()
        
    def handle_async(self):
        """
        Download self.url asynchronously and call self.callback with the result.
        """
        self.callback(self.handle())
        
    def handle(self):
        """
        Download self.url and return the result.
        """
        return self.response_builder('error:no downloader')


class Urllib2Downloader(BaseDownloader):
    """
    A threaded urllib2 downloader, this is the default fallback on non-unix
    systems or unix systems without wget
    """
    import threading
    import functools
    import urllib2

    def handle_async(self):
        def downloader():
            try:
                u = self.urllib2.urlopen(self.url)
            except self.urllib2.URLError:
                callback([])
            self.callback(self.response_builder(u.read()))
        self.threading.Thread(target=self.functools.partial(downloader)).start()

    def handle(self):
        try:
            u = self.urllib2.urlopen(self.url)
        except self.urllib2.URLError:
            callback([])
        return self.response_builder(u.read())

    
class WgetDownloader(BaseDownloader):
    """
    Downloader using the wget application. This runs the download in another
    process and thus creates minimal lag. The downloader checks periodically
    using gamethread whether the download finished or not.
    """
    try:
        import gamethread
        _debug = False
    except ImportError:
        _debug = True
    import os
    tmp_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'lotterylib_tmp_file')
    
    def __init__(self, *args, **kwargs):
        BaseDownloader.__init__(self, *args, **kwargs)
        if self.os.path.exists(self.tmp_file):
            self.os.remove(self.tmp_file)
        self.timer = 0
        
    def handle_async(self):
        self.os.system('wget -q --output-document=%s %s' % (self.tmp_file, self.url))
        self.schedule()
        
    def schedule(self):
        if self._debug:
            import time
            time.sleep(2)
            self.check()
        else:
            self.gamethread.delayed(2, self.check)
        
    def check(self):
        if self.os.path.exists(self.tmp_file):
            f = open(self.tmp_file,'rb')
            c = f.read()
            f.close()
            self.callback(self.response_builder(c))
        else:
            self.timer += 1
            if self.timer < 30:
                self.schedule()
            else:
                self.callback(self.response_builder('error:connection timeout'))
                
    def handle(self):
        pipe = self.os.popen('wget -q -O - %s' % self.url)
        response = pipe.read()
        return self.response_builder(response)


#===============================================================================
# Response
#===============================================================================

class ResponseError(Exception):
    """
    Error which occurs if success-only methods are used in an unsuccessful 
    Response object
    """


class Response(object):
    """
    Since lotterylib encourages usage of asynchronous downloads, we can't just
    raise Exceptions. Thus any download (asynchronous or synchronous) will get a
    Response object.
    """
    def __init__(self, result={}, errmsg=''):
        self.success = not errmsg
        self.error = bool(errmsg)
        self.result = result
        self.errmsg = errmsg
        
    def __iter__(self):
        if self.error:
            self.raise_error()
        else:
            return self.result.iteritems()
        
    def __bool__(self):
        return self.success
    
    def __getitem__(self, item):
        if self.error:
            self.raise_error()
        else:
            return self.result.__getitem__(item)
    
    def __contains__(self, item):
        if self.error:
            self.raise_error()
        else:
            return item in self.result
        
    def raise_error(self):
        raise ResponseError, self.errmsg
    
class ResponseV1(Response):
    """
    Response objects for version 1 of the API protocol.
    """
    
RESPONSE_BUILDERS = {}    

def get_response_builder(version):
    return RESPONSE_BUILDERS[version]


def build_response_1(content):
    """
    Build a Response object.
    """
    if not content:
        return ResponseV1(errmsg='No Response')
    lines = content.split('\n')
    header = lines.pop(0)
    if not header:
        return ResponseV1(errmsg='No Header')
    if header.startswith('error:'):
        return ResponseV1(errmsg=header[6:])
    if not header == 'ok':
        return ResponseV1(errmsg='Unexpected Response: %s' % header)
    result = {}
    current = None
    for line in lines:
        if line.startswith('@'):
            current = line[1:]
            result[current] = []
        elif current:
            result[current].append(line)
    return ResponseV1(result=result)
RESPONSE_BUILDERS[1] = build_response_1

#===============================================================================
# Internal
#===============================================================================

def get_default_downloader():
    """
    On Linux it returns the wget_downloader if wget is installed, otherwise it
    returns the urllib2_downloader
    """
    import os
    if os.name == 'posix':
        if os.popen('wget -V').read().lower().startswith('gnu wget'):
            return WgetDownloader
    return Urllib2Downloader

class Manager(object):
    def __init__(self):
        self.downloader = None
        
    def get_downloader(self):
        if self.downloader:
            return self.downloader
        self.downloader = get_default_downloader()
        return self.downloader
    
    def set_downloader(self, downloader):
        self.downloader = downloader
        
manager = Manager()
get_downloader = manager.get_downloader